// CODE WRITTEN BY MARC S. SHERMAN 2011-2013. marc.s.sherman@gmail.com
// marc.s.sherman@gmail.com or marc.sherman@wustl.edu
// Version 3.0



#include <cmath>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <ctime>
#include <vector>

#include "matrix.h"
#include "analytic.h"
#include "global.h"

using namespace std;

inline
double logrand(double min, double max)
{
	double lmin = log(min);
	double lmax = log(max);

	return exp(rand()*(lmax-lmin)/RAND_MAX + lmin);
}
bool fexists(string filename)
{
  ifstream ifile(filename.c_str());
  return ifile;
}

inline
vector <double> logstepper(int numSteps,double minVal,double maxVal)
{
    vector <double> outVec(numSteps);
    if (numSteps==1)
    {
        outVec[0]=exp((log(maxVal)+log(minVal))/2);
    }
    else
    {
        double stepSize=(log(maxVal)-log(minVal))/(numSteps-1);

        for (int i=0;i<numSteps;i++) outVec[i]=exp(log(minVal)+stepSize*(i));
    }

    return outVec;
}

int main(int argc, char* argv[])
{
	//|| OPTIONS ||||||||||||||||||//
	int default_meshdivisions=50; // resolution of first parameter value in log steps
	int initParamLength=1000; // initial size of matrix to store parameters
	double default_ERR=.001;
	string filename = "out_aco";

	//|| END OPTIONS ||||||||||||||||||//

	Moments protein_moments;

	bool randomize_parameters=false;
	bool get_moments_from_parameters=true;
	double cand_o, cand_f, cand_d, cand_c, cand_k, cand_j;
	int known_mode=0;

	int meshdivisions;
	double ERR;

	if (argc==4)
	{
		meshdivisions=atoi(argv[2]);
		ERR=atof(argv[3]);
		randomize_parameters=true;
	}
	else if (argc==10)
	{
		meshdivisions=atoi(argv[2]);
		ERR=atof(argv[3]);
		cand_o=atof(argv[4]);
		cand_f=atof(argv[5]);
		cand_k=atof(argv[6]);
		cand_d=atof(argv[7]);
		cand_j=atof(argv[8]);
		cand_c=atof(argv[9]);

		randomize_parameters=false;
		get_moments_from_parameters=true;

    }
    else if (argc==11)
	{
		meshdivisions=atoi(argv[2]);
		ERR=atof(argv[3]);
		cand_o=atof(argv[4]);
		cand_f=atof(argv[5]);
		cand_k=atof(argv[6]);
		cand_d=atof(argv[7]);
		cand_j=atof(argv[8]);
		cand_c=atof(argv[9]);

		known_mode=atoi(argv[10]);

		randomize_parameters=false;
		get_moments_from_parameters=true;


    }
	else if (argc==14)
	{
		meshdivisions=atoi(argv[2]);
		ERR=atof(argv[3]);
		randomize_parameters=false;
		cand_o=atof(argv[4]);
		cand_f=atof(argv[5]);
		cand_k=atof(argv[6]);
		cand_d=atof(argv[7]);
		cand_j=atof(argv[8]);
		cand_c=atof(argv[9]);
		protein_moments.mean=atof(argv[10]);
		protein_moments.var=atof(argv[11]);
		protein_moments.skew=atof(argv[12]);
		protein_moments.kurt=atof(argv[13]);

		known_mode=atoi(argv[10]);

		randomize_parameters=false;
		get_moments_from_parameters=false;
	}
	else
	{
		meshdivisions=default_meshdivisions;
		ERR=default_ERR;
	}

	int rand_arg = atoi(argv[1]);



	double on = .001;
	double ox = 50;

	double fn = .001;
	double fx = 50;

	double dn = .000628;
	double dx = 3.1;

	double cn = .00045;
	double cx = 1;

	double kn = .01;
	double kx = 200;

	double jn = .5;
	double jx = 55;

	//Pre-compute the log ranges of these parameters.
	double log_o_range=log(ox/on);
    double log_f_range=log(fx/fn);
    double log_d_range=log(dx/dn);
    double log_c_range=log(cx/cn);
    double log_k_range=log(kx/kn);
    double log_j_range=log(jx/jn);

	// ||| RANDOMLY GENERATE A PARAMETER SET |||
	// RANDOMIZE OPTIONS
	double protein_min = 25;
	double protein_max = 150000;
	// END RAND OPTIONS

	srand(time(NULL)+rand_arg*1000);

	if (randomize_parameters==true)
	{
		do
		{
			cand_o=logrand(on,ox);
			cand_f=logrand(fn,fx);
			cand_d=logrand(dn,dx);
			cand_c=logrand(cn,cx);
			cand_k=logrand(kn,kx);
			cand_j=logrand(jn,jx);
			protein_moments.mean=cme_solution_mean(cand_o,cand_f,cand_d,cand_c,cand_k,cand_j);
		} while (protein_moments.mean > protein_max || protein_moments.mean < protein_min);
	}

	if (get_moments_from_parameters==true)
	{
		protein_moments.mean=cme_solution_mean(cand_o,cand_f,cand_d,cand_c,cand_k,cand_j);
		protein_moments.var=cme_solution_var(cand_o,cand_f,cand_d,cand_c,cand_k,cand_j);
		protein_moments.skew=cme_solution_skew(cand_o,cand_f,cand_d,cand_c,cand_k,cand_j);
		protein_moments.kurt=cme_solution_kurt(cand_o,cand_f,cand_d,cand_c,cand_k,cand_j);
	}

	vector <double> k, j, d, c, o, f;

	double Vtestn, Vtestx, Stestn, Stestx, Ktestn, Ktestx;
	double outVar, outSkew, outKurt;
	double kc, dc, jc, cc, oc, fc;
	double dLB, dUB, cLB, cUB, jLB, jUB, oLB, oUB, kLB, kUB, fLB, fUB;
	double jRange, dRange, cRange, oRange, kRange, fRange;
	double pstar;
	int meshj, meshd, meshc, mesho, meshk, meshf;
	int i1, i2, i3, i4, i5;

	int counter=0;
	int buffCt=0;
	int numParamRows=initParamLength*pow(2,buffCt);
	int oldParamRows;

	Matrix <double> params(numParamRows,6);

	//determine boundaries for moments (relative error)
	ERR=ERR+1;
    Vtestn=protein_moments.var/ERR;
    Vtestx=protein_moments.var*ERR;
    Stestn=protein_moments.skew/ERR;
    Stestx=protein_moments.skew*ERR;
    Ktestn=protein_moments.kurt/ERR;
    Ktestx=protein_moments.kurt*ERR;

    // since skew can be negative, invert the boundaries. otherwise s_min= -4.95 and s_max=-5.05 or something
    if (protein_moments.skew<0)
    {
        double tempS;
        tempS=Stestn;
        Stestn=Stestx;
        Stestx=tempS;
    }

    // check mode. I've rearranged the order here so dm, dp are first.
    //  mode=1 : dp is known
    //  mode=2 : dm is known
    //  mode=3 : both are known.
    //
    // by default, known_mode=0. so none will trigger normally

    if (known_mode==1)
    {
        cc=cand_c;

        // calculate the new upper and lower bounds for k
        kUB=(protein_moments.mean*cc*dx*(on+fx))/(jn*on);
        kLB=(protein_moments.mean*cc*dn*(ox+fn))/(jx*ox);

        //compare the analytical UB/LB with absolute range of k
        if (kUB>kx) kUB=kx;
        if (kLB<kn) kLB=kn;

        if (kUB>kLB)
        {
            //determine how large/small space of j we're now looking at
            kRange=log(kUB/kLB)/log_k_range;

            meshk=ceil(kRange*meshdivisions);
            k=logstepper(meshk,kLB,kUB);

            for (i1=1; i1<=meshdivisions;i1++)
            {
                kc=k[i1-1];

                // calculate the new upper and lower bounds for j
                jUB=(protein_moments.mean*cc*dx*(on+fx))/(kc*on);
                jLB=(protein_moments.mean*cc*dn*(ox+fn))/(kc*ox);

                //compare the analytical UB/LB with absolute range of j
                if (jUB>jx) jUB=jx;
                if (jLB<jn) jLB=jn;

                if (jUB>jLB)
                {
                    //determine how large/small space of j we're now looking at
                    jRange=log(jUB/jLB)/log_j_range;

                    meshj=ceil(jRange*meshdivisions);
                    j=logstepper(meshj,jLB,jUB);

                    for (i2=1; i2<=meshj; i2++)
                    {
                        jc=j[i2-1];

                        // calculate the new upper and lower bounds
                        dUB=(kc*jc*ox)/((ox+fn)*cc*protein_moments.mean);
                        dLB=(kc*jc*on)/((on+fx)*cc*protein_moments.mean);

                        //compare the analytical UB/LB with absolute range
                        if (dUB>dx) dUB=dx;
                        if (dLB<dn) dLB=dn;

                        if (dUB>dLB)
                        {
                            //determine how large/small space
                            dRange=log(dUB/dLB)/log_d_range;

                            //use this number to scale the resolution size.
                            meshd=ceil(dRange*meshdivisions);
                            d=logstepper(meshd,dLB,dUB);

                            for (i3=1; i3<=meshd; i3++)
                            {
                                dc=d[i3-1];

                                //for ton/toff (o, f), ranges of other parameters are irrelevant.

                                pstar=(protein_moments.mean*cc*dc)/(jc*kc);

                                // calculate the new upper and lower bounds
                                oLB=fn*(pstar)/(1-pstar);
                                oUB=fx*(pstar)/(1-pstar);

                                //compare the analytical UB/LB with absolute range
                                if (oUB>ox) oUB=ox;
                                if (oLB<on) oLB=on;

                                if (oUB>oLB)
                                {
                                    //determine how large/small space
                                    oRange=log(oUB/oLB)/log_o_range;

                                    //use this number to scale the resolution size.
                                    mesho=ceil(oRange*meshdivisions);
                                    o=logstepper(mesho,oLB,oUB);

                                    for (i5=1; i5<=mesho; i5++)
                                    {
                                        oc=o[i5-1];
                                        fc=oc*(1-pstar)/pstar;

                                        outVar=cme_solution_var(oc,fc,dc,cc,kc,jc);

                                        if (Vtestn<outVar && Vtestx>outVar)
                                        {
                                            outSkew=cme_solution_skew(oc,fc,dc,cc,kc,jc);

                                            if (Stestn<outSkew && Stestx>outSkew)
                                            {
                                                outKurt=cme_solution_kurt(oc,fc,dc,cc,kc,jc);

                                                if (Ktestn<outKurt && Ktestx>outKurt)
                                                {
                                                    if (counter>=numParamRows)
                                                    {   //resize matrix by adding buffersize
                                                        oldParamRows=numParamRows;

                                                        buffCt=buffCt+1;
                                                        numParamRows=initParamLength*pow(2,buffCt);
                                                        Matrix <double> tempParams(numParamRows,6);
                                                        for (int mm=0;mm<oldParamRows;mm++)
                                                        {
                                                            for (int nn=0;nn<6;nn++)
                                                            {
                                                                tempParams[mm][nn]=params[mm][nn];
                                                            }
                                                        }
                                                        params=tempParams;
                                                    }
                                                    params[counter][0]=oc;
                                                    params[counter][1]=fc;
                                                    params[counter][2]=kc;
                                                    params[counter][3]=dc;
                                                    params[counter][4]=jc;
                                                    params[counter][5]=cc;
                                                    counter++;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (known_mode==2)
    {
        dc=cand_d;

        // calculate the new upper and lower bounds for k
        kUB=(protein_moments.mean*cx*dc*(on+fx))/(jn*on);
        kLB=(protein_moments.mean*cn*dc*(ox+fn))/(jx*ox);

        //compare the analytical UB/LB with absolute range of k
        if (kUB>kx) kUB=kx;
        if (kLB<kn) kLB=kn;

        if (kUB>kLB)
        {
            //determine how large/small space of j we're now looking at
            kRange=log(kUB/kLB)/log_k_range;

            meshk=ceil(kRange*meshdivisions);
            k=logstepper(meshk,kLB,kUB);

            for (i1=1; i1<=meshdivisions;i1++)
            {
                kc=k[i1-1];

                // calculate the new upper and lower bounds for j
                jUB=(protein_moments.mean*cx*dc*(on+fx))/(kc*on);
                jLB=(protein_moments.mean*cn*dc*(ox+fn))/(kc*ox);

                //compare the analytical UB/LB with absolute range of j
                if (jUB>jx) jUB=jx;
                if (jLB<jn) jLB=jn;

                if (jUB>jLB)
                {
                    //determine how large/small space of j we're now looking at
                    jRange=log(jUB/jLB)/log_j_range;

                    meshj=ceil(jRange*meshdivisions);
                    j=logstepper(meshj,jLB,jUB);

                    for (i2=1; i2<=meshj; i2++)
                    {
                        jc=j[i2-1];

                        // calculate the new upper and lower bounds
                        cUB=(kc*jc*ox)/((ox+fn)*dc*protein_moments.mean);
                        cLB=(kc*jc*on)/((on+fx)*dc*protein_moments.mean);

                        //compare the analytical UB/LB with absolute range
                        if (cUB>cx) cUB=cx;
                        if (cLB<cn) cLB=cn;

                        if (cUB>cLB)
                        {
                            //determine how large/small space
                            cRange=log(cUB/cLB)/log_c_range;

                            //use this number to scale the resolution size.
                            meshc=ceil(cRange*meshdivisions);
                            c=logstepper(meshc,cLB,cUB);

                            for (i4=1; i4<=meshc; i4++)
                            {
                                cc=c[i4-1];

                                //for ton/toff (o, f), ranges of other parameters are irrelevant.
                                pstar=(protein_moments.mean*cc*dc)/(jc*kc);

                                // calculate the new upper and lower bounds
                                oLB=fn*(pstar)/(1-pstar);
                                oUB=fx*(pstar)/(1-pstar);

                                //compare the analytical UB/LB with absolute range
                                if (oUB>ox) oUB=ox;
                                if (oLB<on) oLB=on;

                                if (oUB>oLB)
                                {
                                    //determine how large/small space
                                    oRange=log(oUB/oLB)/log_o_range;

                                    //use this number to scale the resolution size.
                                    mesho=ceil(oRange*meshdivisions);
                                    o=logstepper(mesho,oLB,oUB);

                                    for (i5=1; i5<=mesho; i5++)
                                    {
                                        oc=o[i5-1];
                                        fc=oc*(1-pstar)/pstar;

                                        outVar=cme_solution_var(oc,fc,dc,cc,kc,jc);

                                        if (Vtestn<outVar && Vtestx>outVar)
                                        {
                                            outSkew=cme_solution_skew(oc,fc,dc,cc,kc,jc);

                                            if (Stestn<outSkew && Stestx>outSkew)
                                            {
                                                outKurt=cme_solution_kurt(oc,fc,dc,cc,kc,jc);

                                                if (Ktestn<outKurt && Ktestx>outKurt)
                                                {
                                                    if (counter>=numParamRows)
                                                    {   //resize matrix by adding buffersize
                                                        oldParamRows=numParamRows;

                                                        buffCt=buffCt+1;
                                                        numParamRows=initParamLength*pow(2,buffCt);
                                                        Matrix <double> tempParams(numParamRows,6);
                                                        for (int mm=0;mm<oldParamRows;mm++)
                                                        {
                                                            for (int nn=0;nn<6;nn++)
                                                            {
                                                                tempParams[mm][nn]=params[mm][nn];
                                                            }
                                                        }
                                                        params=tempParams;
                                                    }
                                                    params[counter][0]=oc;
                                                    params[counter][1]=fc;
                                                    params[counter][2]=kc;
                                                    params[counter][3]=dc;
                                                    params[counter][4]=jc;
                                                    params[counter][5]=cc;
                                                    counter++;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (known_mode==3)
    {
        dc=cand_d;
        cc=cand_c;

        // calculate the new upper and lower bounds for k
        kUB=(protein_moments.mean*cc*dc*(on+fx))/(jn*on);
        kLB=(protein_moments.mean*cc*dc*(ox+fn))/(jx*ox);

        //compare the analytical UB/LB with absolute range of k
        if (kUB>kx) kUB=kx;
        if (kLB<kn) kLB=kn;

        if (kUB>kLB)
        {
            //determine how large/small space of j we're now looking at
            kRange=log(kUB/kLB)/log_k_range;

            meshk=ceil(kRange*meshdivisions);
            k=logstepper(meshk,kLB,kUB);

            for (i1=1; i1<=meshdivisions;i1++)
            {
                kc=k[i1-1];

                // calculate the new upper and lower bounds for j
                jUB=(protein_moments.mean*cc*dc*(on+fx))/(kc*on);
                jLB=(protein_moments.mean*cc*dc*(ox+fn))/(kc*ox);

                //compare the analytical UB/LB with absolute range of j
                if (jUB>jx) jUB=jx;
                if (jLB<jn) jLB=jn;

                if (jUB>jLB)
                {
                    //determine how large/small space of j we're now looking at
                    jRange=log(jUB/jLB)/log_j_range;

                    meshj=ceil(jRange*meshdivisions);
                    j=logstepper(meshj,jLB,jUB);

                    for (i2=1; i2<=meshj; i2++)
                    {
                        jc=j[i2-1];

                        //for ton/toff (o, f), ranges of other parameters are irrelevant.

                        pstar=(protein_moments.mean*cc*dc)/(jc*kc);

                        // calculate the new upper and lower bounds
                        oLB=fn*(pstar)/(1-pstar);
                        oUB=fx*(pstar)/(1-pstar);

                        //compare the analytical UB/LB with absolute range
                        if (oUB>ox) oUB=ox;
                        if (oLB<on) oLB=on;

                        if (oUB>oLB)
                        {
                            //determine how large/small space
                            oRange=log(oUB/oLB)/log_o_range;

                            //use this number to scale the resolution size.
                            mesho=ceil(oRange*meshdivisions);
                            mesho=max(mesho,1);
                            o=logstepper(mesho,oLB,oUB);

                            for (i5=1; i5<=mesho; i5++)
                            {	oc=o[i5-1];

                                fc=oc*(1-pstar)/pstar;

                                outVar=cme_solution_var(oc,fc,dc,cc,kc,jc);

                                if (Vtestn<outVar && Vtestx>outVar)
                                {
                                    outSkew=cme_solution_skew(oc,fc,dc,cc,kc,jc);

                                    if (Stestn<outSkew && Stestx>outSkew)
                                    {
                                        outKurt=cme_solution_kurt(oc,fc,dc,cc,kc,jc);

                                        if (Ktestn<outKurt && Ktestx>outKurt)
                                        {
                                            if (counter>=numParamRows)
                                            {   //resize matrix by adding buffersize
                                                oldParamRows=numParamRows;

                                                buffCt=buffCt+1;
                                                numParamRows=initParamLength*pow(2,buffCt);
                                                Matrix <double> tempParams(numParamRows,6);
                                                for (int mm=0;mm<oldParamRows;mm++)
                                                {
                                                    for (int nn=0;nn<6;nn++)
                                                    {
                                                        tempParams[mm][nn]=params[mm][nn];
                                                    }
                                                }
                                                params=tempParams;
                                            }
                                            params[counter][0]=oc;
                                            params[counter][1]=fc;
                                            params[counter][2]=kc;
                                            params[counter][3]=dc;
                                            params[counter][4]=jc;
                                            params[counter][5]=cc;
                                            counter++;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (known_mode==4) //check ton/toff first.
    {
        double pstarn=0;
        double pstarx=0;

        pstarn=protein_moments.mean*cn*dn/(jx*kx);
        pstarx=protein_moments.mean*cx*dx/(jn*kn);

        o.resize(meshdivisions);
        o=logstepper(meshdivisions,on,ox);

        for (i1=1; i1<=meshdivisions;i1++)
        {
            oc=o[i1-1];

            fLB=oc*(1-pstarx)/pstarx;
            fUB=oc*(1-pstarn)/pstarn;

            //compare the analytical UB/LB with absolute range of j
            if (fUB>fx) fUB=fx;
            if (fLB<fn) fLB=fn;

            //ensure there are valid solutions given the information so far.
            if (fUB>fLB)
            {
                //determine how large/small space of j we're now looking at
                fRange=log(fUB/fLB)/log_f_range;

                //use this number to scale the resolution size.
                meshf=ceil(fRange*meshdivisions);
                f=logstepper(meshf,fLB,fUB);

                for (i2=1; i2<=meshf; i2++)
                {
                    fc=f[i2-1];

                    // calculate the new upper and lower bounds
                    dUB=(kx*jx*oc)/((oc+fc)*cn*protein_moments.mean);
                    dLB=(kn*jn*oc)/((oc+fc)*cx*protein_moments.mean);

                    //compare the analytical UB/LB with absolute range
                    if (dUB>dx) dUB=dx;
                    if (dLB<dn) dLB=dn;

                    //ensure there are valid solutions given the information so far.
                    if (dUB>dLB)
                    {
                        //determine how large/small space
                        dRange=log(dUB/dLB)/log_d_range;

                        //use this number to scale the resolution size.
                        meshd=ceil(dRange*meshdivisions);
                        d=logstepper(meshd,dLB,dUB);

                        for (i3=1; i3<=meshd; i3++)
                        {

                            dc=d[i3-1];

                            // calculate the new upper and lower bounds
                            cUB=(kx*jx*oc)/((oc+fc)*dc*protein_moments.mean);
                            cLB=(kn*jn*oc)/((oc+fc)*dc*protein_moments.mean);

                            //compare the analytical UB/LB with absolute range
                            if (cUB>cx) cUB=cx;
                            if (cLB<cn) cLB=cn;

                            //ensure there are valid solutions given the information so far.
                            if (cUB>cLB)
                            {
                                //determine how large/small space is in log space
                                cRange=log(cUB/cLB)/log_c_range;

                                //use this number to scale the resolution size.
                                meshc=ceil(cRange*meshdivisions);
                                c=logstepper(meshc,cLB,cUB);

                                for (i4=1; i4<=meshc; i4++)
                                {
                                    cc=c[i4-1];

                                    // calculate the new upper and lower bounds
                                    jUB=protein_moments.mean*(oc+fc)*cc*dc/(oc*kn);
                                    jLB=protein_moments.mean*(oc+fc)*cc*dc/(oc*kx);

                                    //compare the analytical UB/LB with absolute range
                                    if (jUB>jx) jUB=jx;
                                    if (jLB<jn) jLB=jn;

                                    //ensure there are valid solutions given the information so far.
                                    if (jUB>jLB)
                                    {
                                        //determine how large/small space
                                        jRange=log(jUB/jLB)/log_j_range;

                                        //use this number to scale the resolution size.
                                        meshj=ceil(jRange*meshdivisions);
                                        j=logstepper(meshj,jLB,jUB);

                                        for (i5=1; i5<=meshj; i5++)
                                        {
                                            jc=j[i5-1];

                                            kc=protein_moments.mean*(oc+fc)*cc*dc/(oc*jc);

                                            outVar=cme_solution_var(oc,fc,dc,cc,kc,jc);

                                            if (Vtestn<outVar && Vtestx>outVar)
                                            {
                                                outSkew=cme_solution_skew(oc,fc,dc,cc,kc,jc);

                                                if (Stestn<outSkew && Stestx>outSkew)
                                                {
                                                    outKurt=cme_solution_kurt(oc,fc,dc,cc,kc,jc);

                                                    if (Ktestn<outKurt && Ktestx>outKurt)
                                                    {
                                                        if (counter>=numParamRows)
                                                        {   //resize matrix by adding buffersize
                                                            oldParamRows=numParamRows;

                                                            buffCt=buffCt+1;
                                                            numParamRows=initParamLength*pow(2,buffCt);
                                                            Matrix <double> tempParams(numParamRows,6);
                                                            for (int mm=0;mm<oldParamRows;mm++)
                                                            {
                                                                for (int nn=0;nn<6;nn++)
                                                                {
                                                                    tempParams[mm][nn]=params[mm][nn];
                                                                }
                                                            }
                                                            params=tempParams;
                                                        }
                                                        params[counter][0]=oc;
                                                        params[counter][1]=fc;
                                                        params[counter][2]=kc;
                                                        params[counter][3]=dc;
                                                        params[counter][4]=jc;
                                                        params[counter][5]=cc;
                                                        counter++;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (known_mode==5) //check ton/toff first, dm known
    {
        dc=cand_d;

        double pstarn=0;
        double pstarx=0;

        pstarn=protein_moments.mean*cn*dc/(jx*kx);
        pstarx=protein_moments.mean*cx*dc/(jn*kn);

        o.resize(meshdivisions);
        o=logstepper(meshdivisions,on,ox);

        for (i1=1; i1<=meshdivisions;i1++)
        {
            oc=o[i1-1];

            fLB=oc*(1-pstarx)/pstarx;
            fUB=oc*(1-pstarn)/pstarn;

            //compare the analytical UB/LB with absolute range of j
            if (fUB>fx) fUB=fx;
            if (fLB<fn) fLB=fn;

            //ensure there are valid solutions given the information so far.
            if (fUB>fLB)
            {
                //determine how large/small space of j we're now looking at
                fRange=log(fUB/fLB)/log_f_range;

                //use this number to scale the resolution size.
                meshf=ceil(fRange*meshdivisions);
                f=logstepper(meshf,fLB,fUB);

                for (i2=1; i2<=meshf; i2++)
                {
                    fc=f[i2-1];

                    // calculate the new upper and lower bounds
                    cUB=(kx*jx*oc)/((oc+fc)*dc*protein_moments.mean);
                    cLB=(kn*jn*oc)/((oc+fc)*dc*protein_moments.mean);

                    //compare the analytical UB/LB with absolute range
                    if (cUB>cx) cUB=cx;
                    if (cLB<cn) cLB=cn;

                    //ensure there are valid solutions given the information so far.
                    if (cUB>cLB)
                    {
                        //determine how large/small space is in log space
                        cRange=log(cUB/cLB)/log_c_range;

                        //use this number to scale the resolution size.
                        meshc=ceil(cRange*meshdivisions);
                        c=logstepper(meshc,cLB,cUB);

                        for (i4=1; i4<=meshc; i4++)
                        {
                            cc=c[i4-1];

                            // calculate the new upper and lower bounds
                            jUB=protein_moments.mean*(oc+fc)*cc*dc/(oc*kn);
                            jLB=protein_moments.mean*(oc+fc)*cc*dc/(oc*kx);

                            //compare the analytical UB/LB with absolute range
                            if (jUB>jx) jUB=jx;
                            if (jLB<jn) jLB=jn;

                            //ensure there are valid solutions given the information so far.
                            if (jUB>jLB)
                            {
                                //determine how large/small space
                                jRange=log(jUB/jLB)/log_j_range;

                                //use this number to scale the resolution size.
                                meshj=ceil(jRange*meshdivisions);
                                j=logstepper(meshj,jLB,jUB);

                                for (i5=1; i5<=meshj; i5++)
                                {
                                    jc=j[i5-1];

                                    kc=protein_moments.mean*(oc+fc)*cc*dc/(oc*jc);

                                    outVar=cme_solution_var(oc,fc,dc,cc,kc,jc);

                                    if (Vtestn<outVar && Vtestx>outVar)
                                    {
                                        outSkew=cme_solution_skew(oc,fc,dc,cc,kc,jc);

                                        if (Stestn<outSkew && Stestx>outSkew)
                                        {
                                            outKurt=cme_solution_kurt(oc,fc,dc,cc,kc,jc);

                                            if (Ktestn<outKurt && Ktestx>outKurt)
                                            {
                                                if (counter>=numParamRows)
                                                {   //resize matrix by adding buffersize
                                                    oldParamRows=numParamRows;

                                                    buffCt=buffCt+1;
                                                    numParamRows=initParamLength*pow(2,buffCt);
                                                    Matrix <double> tempParams(numParamRows,6);
                                                    for (int mm=0;mm<oldParamRows;mm++)
                                                    {
                                                        for (int nn=0;nn<6;nn++)
                                                        {
                                                            tempParams[mm][nn]=params[mm][nn];
                                                        }
                                                    }
                                                    params=tempParams;
                                                }
                                                params[counter][0]=oc;
                                                params[counter][1]=fc;
                                                params[counter][2]=kc;
                                                params[counter][3]=dc;
                                                params[counter][4]=jc;
                                                params[counter][5]=cc;
                                                counter++;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (known_mode==6) //check ton/toff first, dp known
    {
        cc=cand_c;
        double pstarn=0;
        double pstarx=0;

        pstarn=protein_moments.mean*cc*dn/(jx*kx);
        pstarx=protein_moments.mean*cc*dx/(jn*kn);

        o.resize(meshdivisions);
        o=logstepper(meshdivisions,on,ox);

        for (i1=1; i1<=meshdivisions;i1++)
        {
            oc=o[i1-1];

            fLB=oc*(1-pstarx)/pstarx;
            fUB=oc*(1-pstarn)/pstarn;

            //compare the analytical UB/LB with absolute range of j
            if (fUB>fx) fUB=fx;
            if (fLB<fn) fLB=fn;

            //ensure there are valid solutions given the information so far.
            if (fUB>fLB)
            {
                //determine how large/small space of j we're now looking at
                fRange=log(fUB/fLB)/log_f_range;

                //use this number to scale the resolution size.
                meshf=ceil(fRange*meshdivisions);
                f=logstepper(meshf,fLB,fUB);

                for (i2=1; i2<=meshf; i2++)
                {
                    fc=f[i2-1];

                    // calculate the new upper and lower bounds
                    dUB=(kx*jx*oc)/((oc+fc)*cc*protein_moments.mean);
                    dLB=(kn*jn*oc)/((oc+fc)*cc*protein_moments.mean);

                    //compare the analytical UB/LB with absolute range
                    if (dUB>dx) dUB=dx;
                    if (dLB<dn) dLB=dn;

                    //ensure there are valid solutions given the information so far.
                    if (dUB>dLB)
                    {
                        //determine how large/small space
                        dRange=log(dUB/dLB)/log_d_range;

                        //use this number to scale the resolution size.
                        meshd=ceil(dRange*meshdivisions);
                        d=logstepper(meshd,dLB,dUB);

                        for (i3=1; i3<=meshd; i3++)
                        {
                            dc=d[i3-1];

                            jUB=protein_moments.mean*(oc+fc)*cc*dc/(oc*kn);
                            jLB=protein_moments.mean*(oc+fc)*cc*dc/(oc*kx);

                            //compare the analytical UB/LB with absolute range
                            if (jUB>jx) jUB=jx;
                            if (jLB<jn) jLB=jn;

                            //ensure there are valid solutions given the information so far.
                            if (jUB>jLB)
                            {
                                //determine how large/small space
                                jRange=log(jUB/jLB)/log_j_range;

                                //use this number to scale the resolution size.
                                meshj=ceil(jRange*meshdivisions);
                                j=logstepper(meshj,jLB,jUB);

                                for (i5=1; i5<=meshj; i5++)
                                {
                                    jc=j[i5-1];

                                    kc=protein_moments.mean*(oc+fc)*cc*dc/(oc*jc);

                                    outVar=cme_solution_var(oc,fc,dc,cc,kc,jc);

                                    if (Vtestn<outVar && Vtestx>outVar)
                                    {
                                        outSkew=cme_solution_skew(oc,fc,dc,cc,kc,jc);

                                        if (Stestn<outSkew && Stestx>outSkew)
                                        {
                                            outKurt=cme_solution_kurt(oc,fc,dc,cc,kc,jc);

                                            if (Ktestn<outKurt && Ktestx>outKurt)
                                            {
                                                if (counter>=numParamRows)
                                                {   //resize matrix by adding buffersize
                                                    oldParamRows=numParamRows;

                                                    buffCt=buffCt+1;
                                                    numParamRows=initParamLength*pow(2,buffCt);
                                                    Matrix <double> tempParams(numParamRows,6);
                                                    for (int mm=0;mm<oldParamRows;mm++)
                                                    {
                                                        for (int nn=0;nn<6;nn++)
                                                        {
                                                            tempParams[mm][nn]=params[mm][nn];
                                                        }
                                                    }
                                                    params=tempParams;
                                                }
                                                params[counter][0]=oc;
                                                params[counter][1]=fc;
                                                params[counter][2]=kc;
                                                params[counter][3]=dc;
                                                params[counter][4]=jc;
                                                params[counter][5]=cc;
                                                counter++;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (known_mode==7) //check ton/toff first, dp&dm known
    {
        cc=cand_c;
        dc=cand_d;

        double pstarn=0;
        double pstarx=0;

        pstarn=protein_moments.mean*cc*dc/(jx*kx);
        pstarx=protein_moments.mean*cc*dc/(jn*kn);

        o.resize(meshdivisions);
        o=logstepper(meshdivisions,on,ox);

        for (i1=1; i1<=meshdivisions;i1++)
        {
            oc=o[i1-1];

            fLB=oc*(1-pstarx)/pstarx;
            fUB=oc*(1-pstarn)/pstarn;

            //compare the analytical UB/LB with absolute range of j
            if (fUB>fx) fUB=fx;
            if (fLB<fn) fLB=fn;

            //ensure there are valid solutions given the information so far.
            if (fUB>fLB)
            {
                //determine how large/small space of j we're now looking at
                fRange=log(fUB/fLB)/log_f_range;

                //use this number to scale the resolution size.
                meshf=ceil(fRange*meshdivisions);
                f=logstepper(meshf,fLB,fUB);

                for (i2=1; i2<=meshf; i2++)
                {
                    fc=f[i2-1];

                    // calculate the new upper and lower bounds
                    jUB=protein_moments.mean*(oc+fc)*cc*dc/(oc*kn);
                    jLB=protein_moments.mean*(oc+fc)*cc*dc/(oc*kx);

                    //compare the analytical UB/LB with absolute range
                    if (jUB>jx) jUB=jx;
                    if (jLB<jn) jLB=jn;

                    //ensure there are valid solutions given the information so far.
                    if (jUB>jLB)
                    {
                        //determine how large/small space
                        jRange=log(jUB/jLB)/log_j_range;

                        //use this number to scale the resolution size.
                        meshj=ceil(jRange*meshdivisions);
                        j=logstepper(meshj,jLB,jUB);

                        for (i5=1; i5<=meshj; i5++)
                        {
                            jc=j[i5-1];

                            kc=protein_moments.mean*(oc+fc)*cc*dc/(oc*jc);

                            outVar=cme_solution_var(oc,fc,dc,cc,kc,jc);

                            if (Vtestn<outVar && Vtestx>outVar)
                            {
                                outSkew=cme_solution_skew(oc,fc,dc,cc,kc,jc);

                                if (Stestn<outSkew && Stestx>outSkew)
                                {
                                    outKurt=cme_solution_kurt(oc,fc,dc,cc,kc,jc);

                                    if (Ktestn<outKurt && Ktestx>outKurt)
                                    {
                                        if (counter>=numParamRows)
                                        {   //resize matrix by adding buffersize
                                            oldParamRows=numParamRows;

                                            buffCt=buffCt+1;
                                            numParamRows=initParamLength*pow(2,buffCt);
                                            Matrix <double> tempParams(numParamRows,6);
                                            for (int mm=0;mm<oldParamRows;mm++)
                                            {
                                                for (int nn=0;nn<6;nn++)
                                                {
                                                    tempParams[mm][nn]=params[mm][nn];
                                                }
                                            }
                                            params=tempParams;
                                        }
                                        params[counter][0]=oc;
                                        params[counter][1]=fc;
                                        params[counter][2]=kc;
                                        params[counter][3]=dc;
                                        params[counter][4]=jc;
                                        params[counter][5]=cc;
                                        counter++;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else
    {
        k.resize(meshdivisions);
        k=logstepper(meshdivisions,kn,kx);

        for (i1=1; i1<=meshdivisions;i1++)
        {
            kc=k[i1-1];

            // calculate the new upper and lower bounds for j
            jUB=(protein_moments.mean*cx*dx*(on+fx))/(kc*on);
            jLB=(protein_moments.mean*cn*dn*(ox+fn))/(kc*ox);

            //compare the analytical UB/LB with absolute range of j
            if (jUB>jx) jUB=jx;
            if (jLB<jn) jLB=jn;

            //ensure there are valid solutions given the information so far.
            if (jUB>jLB)
            {
                //determine how large/small space of j we're now looking at
                jRange=log(jUB/jLB)/log_j_range;

                //use this number to scale the resolution size.
                meshj=ceil(jRange*meshdivisions);
                j=logstepper(meshj,jLB,jUB);

                for (i2=1; i2<=meshj; i2++)
                {
                    jc=j[i2-1];

                    // calculate the new upper and lower bounds
                    dUB=(kc*jc*ox)/((ox+fn)*cn*protein_moments.mean);
                    dLB=(kc*jc*on)/((on+fx)*cx*protein_moments.mean);

                    //compare the analytical UB/LB with absolute range
                    if (dUB>dx) dUB=dx;
                    if (dLB<dn) dLB=dn;

                    //ensure there are valid solutions given the information so far.
                    if (dUB>dLB)
                    {
                        //determine how large/small space
                        dRange=log(dUB/dLB)/log_d_range;

                        //use this number to scale the resolution size.
                        meshd=ceil(dRange*meshdivisions);
                        d=logstepper(meshd,dLB,dUB);

                        for (i3=1; i3<=meshd; i3++)
                        {
                            dc=d[i3-1];

                            // calculate the new upper and lower bounds
                            cUB=(kc*jc*ox)/((ox+fn)*dc*protein_moments.mean);
                            cLB=(kc*jc*on)/((on+fx)*dc*protein_moments.mean);

                            //compare the analytical UB/LB with absolute range
                            if (cUB>cx) cUB=cx;
                            if (cLB<cn) cLB=cn;

                            //ensure there are valid solutions given the information so far.
                            if (cUB>cLB)
                            {
                                //determine how large/small space is in log space
                                cRange=log(cUB/cLB)/log_c_range;

                                //use this number to scale the resolution size.
                                meshc=ceil(cRange*meshdivisions);
                                c=logstepper(meshc,cLB,cUB);

                                for (i4=1; i4<=meshc; i4++)
                                {
                                    cc=c[i4-1];

                                    //for ton/toff (o, f), ranges of other parameters are irrelevant.
                                    pstar=(protein_moments.mean*cc*dc)/(jc*kc);

                                    // calculate the new upper and lower bounds
                                    oUB=fx*(pstar)/(1-pstar);
                                    oLB=fn*(pstar)/(1-pstar);

                                    //compare the analytical UB/LB with absolute range
                                    if (oUB>ox) oUB=ox;
                                    if (oLB<on) oLB=on;

                                    //ensure there are valid solutions given the information so far.
                                    if (oUB>oLB)
                                    {
                                        //determine how large/small space
                                        oRange=log(oUB/oLB)/log_o_range;

                                        //use this number to scale the resolution size.
                                        mesho=ceil(oRange*meshdivisions);
                                        o=logstepper(mesho,oLB,oUB);

                                        for (i5=1; i5<=mesho; i5++)
                                        {
                                            oc=o[i5-1];

                                            fc=oc*(1-pstar)/pstar;

                                            outVar=cme_solution_var(oc,fc,dc,cc,kc,jc);

                                            if (Vtestn<outVar && Vtestx>outVar)
                                            {
                                                outSkew=cme_solution_skew(oc,fc,dc,cc,kc,jc);

                                                if (Stestn<outSkew && Stestx>outSkew)
                                                {
                                                    outKurt=cme_solution_kurt(oc,fc,dc,cc,kc,jc);

                                                    if (Ktestn<outKurt && Ktestx>outKurt)
                                                    {
                                                        if (counter>=numParamRows)
                                                        {   //resize matrix by adding buffersize
                                                            oldParamRows=numParamRows;

                                                            buffCt=buffCt+1;
                                                            numParamRows=initParamLength*pow(2,buffCt);
                                                            Matrix <double> tempParams(numParamRows,6);
                                                            for (int mm=0;mm<oldParamRows;mm++)
                                                            {
                                                                for (int nn=0;nn<6;nn++)
                                                                {
                                                                    tempParams[mm][nn]=params[mm][nn];
                                                                }
                                                            }
                                                            params=tempParams;
                                                        }
                                                        params[counter][0]=oc;
                                                        params[counter][1]=fc;
                                                        params[counter][2]=kc;
                                                        params[counter][3]=dc;
                                                        params[counter][4]=jc;
                                                        params[counter][5]=cc;
                                                        counter++;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    string tmpAname=filename;
    string tmpAPname=filename;
	stringstream outID;
	outID << rand_arg;

	tmpAname=tmpAname.append(outID.str());
	tmpAname=tmpAname.append(".txt");

    tmpAPname=tmpAPname.append("_p_");
	tmpAPname=tmpAPname.append(outID.str());
	tmpAPname=tmpAPname.append(".txt");

	ofstream out(tmpAname.c_str());

	for (int a=0; a<counter; a++)
	{	out << params[a][0] << "\t" << params[a][1] << "\t" << params[a][2] << "\t"
		 << params[a][3] << "\t" << params[a][4] << "\t" << params[a][5] << "\n";
	}
	out.close();

	ofstream out_p(tmpAPname.c_str());
	out_p << protein_moments.mean << "\t" << protein_moments.var << "\t" << protein_moments.skew << "\t" << protein_moments.kurt << "\t" << cand_o << "\t" << cand_f << "\t" << cand_k
		 << "\t" << cand_d << "\t" << cand_j << "\t" << cand_c << "\t" << on << "\t" << fn << "\t" << kn
		  << "\t" << dn << "\t" << jn << "\t" << cn << "\t" << ox << "\t" << fx << "\t" << kx << "\t" << dx
		  << "\t" << jx << "\t" << cx;
	out_p.close();


return 0;
}
