//
// MATRIX.CC
// An efficient 2D array class.  Efficiency of access is
// obtained by ensuring that the first dimension of the
// array is internally a power of two.
//
// PhyloNet 3.0
// Copyright (C) 2006-8 by Jeremy Buhler
// All rights reserved

#include <algorithm>

#include "matrix.h"

//
// Matrix constructor
//
template <class EltT> 
Matrix<EltT>::Matrix(unsigned int nx, unsigned int ny)
{
  _nx = nx; _ny = ny;
  
  logLen = computeLogLen(_ny);
  
  values = new EltT [_nx << logLen]; 
}


//
// Matrix::copy()
// common copy code for Matrix copy ctor and operator=
//
template <class EltT> 
void 
Matrix<EltT>::copy(const Matrix &other)
{
  if (values) delete [] values;
  
  _nx = other._nx;
  _ny = other._ny;
  logLen = other.logLen;
  
  values = new EltT [_nx << logLen]; 
  std::copy(other.values, other.values + (_nx << logLen), values);
}


template <class EltT>
void
Matrix<EltT>::fill(EltT v)
{
  for (unsigned int x = 0; x < nx(); x++)
    for (unsigned int y = 0; y < ny(); y++)
      (*this)[x][y] = v;
}
