//
// MATRIX.H
// An efficient 2D array class
//
// PhyloNet 3.0
// Copyright (C) 2006-8 by Jeremy Buhler
// All rights reserved

#ifndef __MATRIX_H
#define __MATRIX_H

template <class EltT> 
class Matrix {
public:
  Matrix(unsigned int nx, unsigned int ny);
  
  Matrix(const Matrix &other)
    : values(0)
  { copy(other); }
  
  Matrix &operator=(const Matrix &other)
  { if (this != &other) copy(other); return *this; }
  
  virtual ~Matrix(void)
  { if (values) delete [] values; }
  
  unsigned int nx() const { return _nx; }
  unsigned int ny() const { return _ny; }
  
  const EltT *operator[](unsigned int r) const
  { return &values[r << logLen]; }

  // non-const row reference
  EltT *operator[](unsigned int r)
  { return &values[r << logLen]; }
      
  void fill(EltT elt);
  
private:
  unsigned int _nx, _ny;
  
  EltT *values;           // elements of the matrix
  unsigned int logLen;    // log(smallest power of 2 >= _ny)
  
  void copy(const Matrix &other);
  
  // compute log of next power of 2 >= vIn.
  static unsigned int computeLogLen(unsigned int vIn)
  {
    unsigned int vLog = (unsigned int) -1;
    for (unsigned int v = vIn; v != 0; v >>= 1, vLog++);
    return ((1U << vLog) == vIn ? vLog : vLog + 1);
  }
};

#include "matrix.cc"

#endif
